use crate::config::Config;
use crate::felinet::{self, PacketRequest};
use crate::felinet::{
    handler_client::HandlerClient, packet_filter::Filter, PacketFilter, PacketIn as SemiPacketIn,
};
use crate::radio::MAX_PACKET_LEN;
use crate::util::{append_internet_to_packet_route, attempt_decode, print_packet};
use anyhow::{anyhow, Context};
use async_stream::stream;
use half::f16;
use ham_cats::buffer::Buffer;
use ham_cats::identity::Identity;
use ham_cats::whisker::NodeInfoBuilder;
use ham_cats::{
    packet::Packet,
    whisker::{Gps, Identification, Route},
};
use std::sync::Arc;
use std::time::{Duration, Instant};
use tokio::sync::{broadcast, mpsc, Mutex};
use tokio_stream::StreamExt;
use tonic::{transport::Channel, Request};
use uuid::Uuid;

const HARDWARE_ID: u16 = 0x7c85;
const SOFTWARE_ID: u8 = 0x00;

pub fn beacon_forever(
    c: &Config,
    modem_send: mpsc::Sender<Vec<u8>>,
    felinet_send: broadcast::Sender<SemiPacketIn>,
    uuid: Uuid,
    temperature_mutex: Arc<Mutex<f32>>,
) -> anyhow::Result<()> {
    let start_time = Instant::now();

    let mut buf = [0; MAX_PACKET_LEN];
    let mut packet = Packet::new(&mut buf);
    packet
        .add_identification(
            Identification::new(c.icon, &c.callsign, c.ssid).context(
                "Config does not describe a valid identification (callsign, ssid, icon)",
            )?,
        )
        .map_err(|e| anyhow!("Could not add identification to beacon packet: {e}"))?;

    let period = c.beacon.period_seconds;

    if let (Some(latitude), Some(longitude), Some(altitude)) =
        (c.beacon.latitude, c.beacon.longitude, c.beacon.altitude)
    {
        packet
            .add_gps(Gps::new(
                latitude,
                longitude,
                f16::from_f64(altitude),
                0,
                0.0,
                f16::ZERO,
            ))
            .map_err(|e| anyhow!("Could not add gps to beacon packet: {e}"))?;
    }

    if let Some(c) = &c.beacon.comment {
        packet
            .add_comment(c)
            .map_err(|e| anyhow!("Could not add gps to beacon packet: {e}"))?;
    }

    let mut info_builder = NodeInfoBuilder::default()
        .hardware_id(HARDWARE_ID)
        .software_id(SOFTWARE_ID);

    if let Some(height) = c.beacon.antenna_height {
        info_builder = info_builder.antenna_height(height);
    }

    if let Some(gain) = c.beacon.antenna_gain {
        info_builder = info_builder.antenna_gain(gain as f64);
    }

    if let Some(power) = c.beacon.tx_power {
        info_builder = info_builder.tx_power(power as f64);
    }

    // This is going to be overwritten.
    // We add it now to ensure there are no errors when we unwrap it later
    packet
        .add_node_info(info_builder.build())
        .map_err(|e| anyhow!("Could not add info to beacon packet: {e}"))?;

    let mut internet_buf = [0; MAX_PACKET_LEN];
    let mut internet_packet = packet.clone_backing(&mut internet_buf);
    let mut rf_packet = packet;

    // internet route
    let mut r = Route::new(c.beacon.max_hops);
    r.push_internet().context("Could not create beacon route")?;
    internet_packet
        .add_route(r)
        .map_err(|e| anyhow!("Could not add route to beacon packet: {e}"))?;

    // rf route
    let r = Route::new(c.beacon.max_hops);
    rf_packet
        .add_route(r)
        .map_err(|e| anyhow!("Could not add route to beacon packet: {e}"))?;

    let rf_packet_bytes = rf_packet.encode().to_vec();
    let internet_packet_bytes = internet_packet.encode().to_vec();
    tokio::task::spawn(async move {
        let mut rf_packet_buf = [0; MAX_PACKET_LEN];
        let mut rf_packet_buf = Buffer::new_empty(&mut rf_packet_buf);
        rf_packet_buf.extend(&rf_packet_bytes);
        let mut rf_packet = Packet::decode(rf_packet_buf).expect("Could not re-create rf packet");

        let mut internet_packet_buf = [0; MAX_PACKET_LEN];
        let mut internet_packet_buf = Buffer::new_empty(&mut internet_packet_buf);
        internet_packet_buf.extend(&internet_packet_bytes);
        let mut internet_packet =
            Packet::decode(internet_packet_buf).expect("Could not re-create internet packet");

        let mut full = [0; MAX_PACKET_LEN];
        let mut tmp = [0; MAX_PACKET_LEN];

        loop {
            tokio::time::sleep(period).await;

            let uptime = (Instant::now() - start_time)
                .as_secs()
                .try_into()
                .expect("Could not calculate uptime");

            let info = info_builder
                .uptime(uptime)
                .xcvr_temperature(
                    temperature_mutex
                        .lock()
                        .await
                        .clamp(i8::MIN as f32, i8::MAX as f32) as i8,
                )
                .build();

            rf_packet.clear_node_info();
            rf_packet
                .add_node_info(info)
                .expect("Could not add node info to rf packet");

            let mut full = Buffer::new_empty(&mut full);
            rf_packet
                .clone_backing(&mut tmp)
                .fully_encode(&mut full)
                .expect("Could not fully encode beacon packet");

            let full = full.to_vec();

            internet_packet.clear_node_info();
            internet_packet
                .add_node_info(info)
                .expect("Could not add node info to rf packet");

            let semi = internet_packet
                .clone_backing(&mut tmp)
                .semi_encode()
                .expect("Could not encode beacon packet")
                .to_vec();
            let semi = SemiPacketIn {
                raw: semi,
                uuid: uuid.into(),
            };

            let _ = modem_send.send(full.clone()).await;
            let _ = felinet_send.send(semi.clone());
        }
    });

    Ok(())
}

pub async fn packet_handler(
    callsign: String,
    ssid: u8,
    tx: broadcast::Sender<SemiPacketIn>,
    packet_send: mpsc::Sender<Vec<u8>>,
    mut packet_receive: mpsc::Receiver<(Vec<u8>, f64)>,
    uuid: Uuid,
) -> anyhow::Result<()> {
    loop {
        let (packet, rssi) = packet_receive
            .recv()
            .await
            .context("Packet receive channel died")?;

        let mut buf = [0; MAX_PACKET_LEN];
        if let Ok(mut packet) = attempt_decode(&packet, &mut buf, rssi) {
            let mut buf2 = [0; MAX_PACKET_LEN];
            let mut internet_packet = packet.clone_backing(&mut buf2);
            match append_internet_to_packet_route(&callsign, ssid, rssi, &mut internet_packet) {
                Some(()) => match internet_packet.semi_encode() {
                    Ok(encoded) => {
                        let semi = SemiPacketIn {
                            raw: encoded.to_vec(),
                            uuid: uuid.into(),
                        };

                        let _ = tx.send(semi);
                    }

                    Err(e) => {
                        eprintln!("Could not semi-encode packet: {e:?}");
                    }
                },
                None => {
                    eprintln!(
                        "Could not add routing to packet before gating to the Internet. Skipping"
                    );
                }
            }

            match packet.should_digipeat(Identity::new(&callsign, ssid)) {
                Ok(()) => {
                    if let Err(e) = packet.append_to_route(&callsign, ssid, Some(rssi)) {
                        eprintln!(
                            "Could not add routing to packet before digipeating: {e}. Skipping"
                        );
                        continue;
                    };

                    let mut encoded_buf = [0; MAX_PACKET_LEN];
                    let mut encoded = Buffer::new_empty(&mut encoded_buf);
                    match packet.fully_encode(&mut encoded) {
                        Ok(()) => {
                            packet_send
                                .send(encoded.to_vec())
                                .await
                                .with_context(|| "Packet send channel died")?;
                        }

                        Err(e) => {
                            eprintln!("Could not fully-encode packet: {e:?}");
                        }
                    }
                }

                Err(e) => {
                    eprintln!("Not digipeating: {e}");
                }
            };
        }
    }
}

pub async fn felinet_send_forever(
    mut client: HandlerClient<Channel>,
    tx: broadcast::Sender<SemiPacketIn>,
) {
    loop {
        let mut rx = tx.subscribe();
        let felinet_stream = stream! {
            loop {
                yield rx.recv().await.expect("Channel closed");
            }
        };

        if let Err(e) = client.push_packets(Request::new(felinet_stream)).await {
            eprintln!("FELINET TX error: {e}. Reconnecting...");
        };

        tokio::time::sleep(Duration::from_secs(1)).await;
    }
}

pub async fn felinet_receive_forever(
    mut client: HandlerClient<Channel>,
    mut modem_send: mpsc::Sender<Vec<u8>>,
    callsign: &str,
    ssid: u8,
    uuid: Uuid,
) {
    loop {
        if let Err(e) =
            felinet_receive_once(&mut client, &mut modem_send, callsign, ssid, uuid).await
        {
            eprintln!("FELINET RX error: {e}. Reconnecting...");
        };
        tokio::time::sleep(Duration::from_secs(1)).await;
    }
}

pub async fn felinet_receive_once(
    client: &mut HandlerClient<Channel>,
    modem_send: &mut mpsc::Sender<Vec<u8>>,
    callsign: &str,
    ssid: u8,
    uuid: Uuid,
) -> anyhow::Result<()> {
    let mut stream = client
        .get_packets(Request::new(PacketRequest {
            filters: vec![
                PacketFilter {
                    filter: Some(Filter::Dest(felinet::CallsignFilter {
                        callsign: callsign.to_string(),
                        ssid: ssid.into(),
                    })),
                },
                // packets directed towards stations we've heard in the last hour
                PacketFilter {
                    filter: Some(Filter::HeardSeconds(3600)),
                },
            ],
            uuid: uuid.into(),
        }))
        .await?
        .into_inner();

    while let Some(semi) = stream.next().await {
        let mut buf = [0; MAX_PACKET_LEN];
        let mut buf = Buffer::new_empty(&mut buf);

        if let Some(pkt) = buf
            .try_extend_from_slice(&semi?.raw[..])
            .ok()
            .and_then(|_| Packet::semi_decode(buf).ok())
        {
            if pkt.should_digipeat(Identity::new(callsign, ssid)).is_ok() {
                print_packet(&pkt, None);

                let mut buf2 = [0; MAX_PACKET_LEN];
                let mut data = Buffer::new_empty(&mut buf2);
                if let Ok(()) = pkt.fully_encode(&mut data) {
                    modem_send.send(data.to_vec()).await?;
                }
            }
        }
    }

    Ok(())
}
